/**
 * This test driver program reads an encoded record from a TAP3 file 
 * and decodes it and displays the results..
 */
#include <stdio.h>
#include <stdlib.h>
#include "TAP-0311.h"
#include "TAP-0312.h"
#include "rtxsrc/rtxStreamFile.h"
#include "rtxsrc/rtxDiag.h"
#include "rtxsrc/rtxMemLeakCheck.h"

int readByVersion (const char* filename, int specVersion, int relVersion)
{  
   OSCTXT	ctxt;
   int		stat;
   OSBOOL     trace = TRUE;

   if ( specVersion != 3 ) {
      printf("Unsupported specification version number: %d", specVersion);
      return -1;
   }
   
   /* Init context for stream access */

   if ((stat = berStrmInitContext (&ctxt)) != 0) {
      printf ("Initialization failed, status %d\n", stat);
      return -1;
   }

   /* Open the input stream to read data */

   stat = rtxStreamFileCreateReader (&ctxt, filename);
   if (stat != 0) {
      rtxErrPrint (&ctxt);
      return stat;
   }

   switch( relVersion ) {
      case 11: {
         TAP_0311_DataInterChange dataInterChange;
         stat = asn1BSD_TAP_0311_DataInterChange (&ctxt, &dataInterChange, ASN1EXPL, 0);

         if (stat == 0) {
            if (trace) {
               printf ("Decode of DataInterChange was successful\n");
               asn1Print_TAP_0311_DataInterChange ("DataInterChange", &dataInterChange);
            }
         }
         else {
            rtxErrPrint (&ctxt); 
            printf ("Decode of DataInterChange failed\n");
            return -1;
         }
         break;
      }
      case 12: {
         TAP_0312_DataInterChange dataInterChange;
         stat = asn1BSD_TAP_0312_DataInterChange (&ctxt, &dataInterChange, 
            ASN1EXPL, 0);

         if (stat == 0) {
            if (trace) {
               printf ("Decode of DataInterChange was successful\n");
               asn1Print_TAP_0312_DataInterChange ("DataInterChange", &dataInterChange);
            }
         }
         else {
            rtxErrPrint (&ctxt); 
            printf ("Decode of DataInterChange failed\n");
            return -1;
         }
         break;
      }
      default: {
         printf("Unsupported release version number: %d",  relVersion);
         return -1;
      }
   }
      
   berStrmFreeContext (&ctxt);

   return 0;
}


int preDecodeVersion_Notification (OSCTXT* pctxt, int length, int* pSpecVersion, int* pRelVersion)
{
   int stat = 0;
   ASN1CCB ccb;
   int len;

   RTDIAGSTRM2 (pctxt,"preDecodeVersion_Notification: start\n");

   len = length;

   /* decode SEQUENCE */

   ccb.len = len;
   ccb.bytes = OSRTSTREAM_BYTEINDEX (pctxt);
   ccb.seqx = 0;

   /* we ignore everything beyond the spec and rel version numbers */
   while (!BS_CHKEND (pctxt, &ccb) && ccb.seqx <=8 ) {
      switch (ccb.seqx) {
      case 0: {
         /* decode sender */
         TAP_0311_Sender sender;
         stat = asn1BSD_TAP_0311_Sender (pctxt, &sender, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 1: {
         /* decode recipient */
         TAP_0311_Recipient recipient;
         stat = asn1BSD_TAP_0311_Recipient (pctxt, &recipient, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 2: {
         /* decode fileSequenceNumber */
         TAP_0311_FileSequenceNumber fileSequenceNumber;
         stat = asn1BSD_TAP_0311_FileSequenceNumber (pctxt, &fileSequenceNumber, 
            ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 3: {
         /* decode rapFileSequenceNumber */
         TAP_0311_RapFileSequenceNumber rapFileSequenceNumber;
         stat = asn1BSD_TAP_0311_RapFileSequenceNumber (pctxt, 
            &rapFileSequenceNumber, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 4: {
         /* decode fileCreationTimeStamp */
         TAP_0311_FileCreationTimeStamp fileCreationTimeStamp;
         stat = asn1BSD_TAP_0311_FileCreationTimeStamp (pctxt, 
            &fileCreationTimeStamp, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 5: {
         /* decode fileAvailableTimeStamp */
         TAP_0311_FileAvailableTimeStamp fileAvailableTimeStamp;
         stat = asn1BSD_TAP_0311_FileAvailableTimeStamp (pctxt, 
            &fileAvailableTimeStamp, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 6: {
         /* decode transferCutOffTimeStamp */
         TAP_0311_TransferCutOffTimeStamp transferCutOffTimeStamp;
         stat = asn1BSD_TAP_0311_TransferCutOffTimeStamp (pctxt, 
            &transferCutOffTimeStamp, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 7: {
         /* decode specificationVersionNumber */
         /* may need to test for value > 0xFFFF */
         stat = asn1BSD_TAP_0311_SpecificationVersionNumber (pctxt, 
            pSpecVersion, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 8: {
         /* decode releaseVersionNumber */
         /* may need to test for value > 0xFFFF */         
         stat = asn1BSD_TAP_0311_ReleaseVersionNumber (pctxt, pRelVersion,
            ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      
      }
   
      if (stat != 0) return LOG_RTERR (pctxt, stat);
      else ccb.seqx++;
   }

   if (ccb.stat != 0) return LOG_RTERR (pctxt, ccb.stat);

   RTDIAGSTRM2 (pctxt,"preDecodeVersion_Notification: end\n");

   return (stat);
}


int preDecodeVersion_BatchControlInfo (OSCTXT* pctxt, int length,
   int* pSpecVersion, int* pRelVersion)
{
   int stat = 0;
   ASN1CCB ccb;
   int len;
   
   RTDIAGSTRM2 (pctxt,"preDecodeVersion_BatchControlInfo: start\n");

   stat = berDecStrmMatchTag (pctxt, TM_APPL|TM_CONS|4, &len, TRUE);
   if (stat != 0) return LOG_RTERR (pctxt, stat);

   /* decode SEQUENCE */   

   ccb.len = len;
   ccb.bytes = OSRTSTREAM_BYTEINDEX (pctxt);
   ccb.seqx = 0;

   /* we ignore everything beyond the spec & version numbers */
   while (!BS_CHKEND (pctxt, &ccb) && ccb.seqx <= 7) {
      switch (ccb.seqx) {
      case 0: {
         /* decode sender */
         TAP_0311_Sender sender;
         stat = asn1BSD_TAP_0311_Sender (pctxt, &sender, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 1: {
         /* decode recipient */
         TAP_0311_Recipient recipient;
         stat = asn1BSD_TAP_0311_Recipient (pctxt, &recipient, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 2: {
         /* decode fileSequenceNumber */
         TAP_0311_FileSequenceNumber fileSequenceNumber;
         stat = asn1BSD_TAP_0311_FileSequenceNumber (pctxt, &fileSequenceNumber, 
            ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 3: {
         /* decode fileCreationTimeStamp */
         TAP_0311_FileCreationTimeStamp fileCreationTimeStamp;
         stat = asn1BSD_TAP_0311_FileCreationTimeStamp (pctxt, 
            &fileCreationTimeStamp, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 4: {
         /* decode transferCutOffTimeStamp */
         TAP_0311_TransferCutOffTimeStamp transferCutOffTimeStamp;
         stat = asn1BSD_TAP_0311_TransferCutOffTimeStamp (pctxt, 
            &transferCutOffTimeStamp, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 5: {
         /* decode fileAvailableTimeStamp */
         TAP_0311_FileAvailableTimeStamp fileAvailableTimeStamp;
         stat = asn1BSD_TAP_0311_FileAvailableTimeStamp (pctxt, 
            &fileAvailableTimeStamp, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 6: {
         /* decode specificationVersionNumber */
         /* may need to test for value > 0xFFFF */
         stat = asn1BSD_TAP_0311_SpecificationVersionNumber (pctxt, 
            pSpecVersion, ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      case 7: {
         /* decode releaseVersionNumber */  
         /* may need to test for value > 0xFFFF */         
         stat = asn1BSD_TAP_0311_ReleaseVersionNumber (pctxt, pRelVersion,
            ASN1EXPL, length);
         if (stat == RTERR_IDNOTFOU) {
            stat = rtxErrReset (pctxt);
         }
         break;
      }
      
      }

      if (stat != 0) return LOG_RTERR (pctxt, stat);
      else ccb.seqx++;
   }

   if (ccb.stat != 0) return LOG_RTERR (pctxt, ccb.stat);

   RTDIAGSTRM2 (pctxt,"preDecodeVersion_BatchControlInfo: end\n");

   return (stat);
}


int preDecodeVersion_TransferBatch (OSCTXT* pctxt,
   int length, int* pSpecVersion, int* pRelVersion)
{
   int stat = 0;
   ASN1CCB ccb;
   int len;

   RTDIAGSTRM2 (pctxt,"preDecodeVersion_TransferBatch: start\n");

   len = length;

   /* decode SEQUENCE */

   ccb.len = len;
   ccb.bytes = OSRTSTREAM_BYTEINDEX (pctxt);
   ccb.seqx = 0;

   if (!BS_CHKEND (pctxt, &ccb)) {
      /* decode batchControlInfo */
      stat = preDecodeVersion_BatchControlInfo (pctxt, length, pSpecVersion, pRelVersion);
      if (stat == RTERR_IDNOTFOU) {
         stat = rtxErrReset (pctxt);
      }

      if (stat != 0) return LOG_RTERR (pctxt, stat);
   }

   if (ccb.stat != 0) return LOG_RTERR (pctxt, ccb.stat);


   RTDIAGSTRM2 (pctxt,"preDecodeVersion_TransferBatch: end\n");

   return (stat);
}


int preDecodeVersion_DataInterChange (OSCTXT* pctxt,
   int length, int* pSpecVersion, int* pRelVersion)
{
   int stat = 0;
   ASN1TAG ctag;
   int len;

   RTDIAGSTRM2 (pctxt,"preDecodeVersion_DataInterChange: start\n");

   len = length;

   /* decode CHOICE */

   rtxStreamMark (pctxt, 32);

   stat = berDecStrmTagAndLen (pctxt, &ctag, &length);
   if (stat != 0) return LOG_RTERR (pctxt, stat);

   switch (ctag)
   {
      case (TM_APPL|TM_CONS|1):
         RTXCTXTPUSHELEMNAME (pctxt, "transferBatch");

         stat = preDecodeVersion_TransferBatch (pctxt, length, pSpecVersion, pRelVersion);
         if (stat != 0) return LOG_RTERR (pctxt, stat);

         RTXCTXTPOPELEMNAME (pctxt);
         break;

      case (TM_APPL|TM_CONS|2):
         RTXCTXTPUSHELEMNAME (pctxt, "notification");

         stat = preDecodeVersion_Notification (pctxt, length, pSpecVersion, pRelVersion);

         if (stat != 0) return LOG_RTERR (pctxt, stat);

         RTXCTXTPOPELEMNAME (pctxt);
         break;

      default:
         rtxStreamReset (pctxt);
   }

   RTDIAGSTRM2 (pctxt,"preDecodeVersion_DataInterChange: end\n");

   return (stat);
}



int main (int argc, char** argv)
{
   
   /* The specification and release version numbers read from input file. */
   int specVersion;
   int relVersion;

   OSCTXT	ctxt;
   int		i, stat;
   const char*  filename = "message.dat";
   OSBOOL     trace = TRUE;

   DECLARE_MEMLEAK_DETECTOR;

   /* Process command line arguments */

   if (argc > 1) {
      for (i = 1; i < argc; i++) {
         if (!strcmp (argv[i], "-v")) rtxSetGlobalDiag (1);
         else if (!strcmp (argv[i], "-i")) filename = argv[++i];
         else if (!strcmp (argv[i], "-notrace")) trace = FALSE;
         else {
            printf ("usage: reader [ -v ] [ -i <filename>\n");
            printf ("   -v  verbose mode: print trace info\n");
            printf ("   -i <filename>  read encoded msg from <filename>\n");
            printf ("   -notrace  do not display trace info\n");
            return 0;
         }
      }
   }

   /* Init context for stream access */

   if ((stat = berStrmInitContext (&ctxt)) != 0) {
      printf ("Initialization failed, status %d\n", stat);
      return -1;
   }

   /* Open the input stream to read data */

   stat = rtxStreamFileCreateReader (&ctxt, filename);
   if (stat != 0) {
      rtxErrPrint (&ctxt);
      return stat;
   }

   /* Determine the spec and release version */

   stat = preDecodeVersion_DataInterChange (&ctxt, 0, &specVersion, &relVersion);

   if (stat != 0) {
      rtxErrPrint (&ctxt); 
      printf ("Version determination for DataInterChange failed\n");
      return -1;
   }

   berStrmFreeContext (&ctxt);
   
   /* Now decode according to the spec and release version */
   return readByVersion(filename, specVersion, relVersion);
}